// Navigation Class
/*
  Encapusulates several popular methods of navigation
  which were clearly ripped off from Dave's examples.
  All you have to do is set the navmode and let it run...
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stream.h>
#include <sys/time.h>

#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfEarthSky.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pf/pfNode.h>
#include <Performer/pf/pfSCS.h>
#include <Performer/pf/pfDCS.h>
#include <Performer/pfdu.h>
#include <pfcave.h>

#include "NavBase.hh"
#include "CavePointer.hh"
#include "ObjectMasks.hh"
//#define DEFAULT_TSPEED 2.0f/1.0e6
//#define DEFAULT_RSPEED 1.0f/1.0e6
#define DEFAULT_TSPEED 30.0f
#define DEFAULT_RSPEED 30.0f
#ifndef PI
#define PI 3.1415
#endif

#define THRESH(val,thresh) (val>thresh || val<-thresh)
//==========>NavBase<====================================
void NavBase::rotate(float rads,char axis){
  CAVENavRot(rads,axis);
}
void NavBase::rotate(float *r){
  rotate(r[0],'x');
  rotate(r[1],'y');
  rotate(r[2],'z');
}
void NavBase::translate(float x,float y,float z){
  CAVENavTranslate(x,y,z);
}
void NavBase::translate(float *v){
  CAVENavTranslate(v[0],v[1],v[2]);
}

void NavBase::loadIdentity(){
  CAVENavLoadIdentity();
}
  
NavBase::NavBase(pfDCS *nav):navDCS(nav){
}

void NavBase::updateDCS() {// do nothing at this level   
  pfCAVEDCSNavTransform(navDCS);
}

void NavBase::update() {// do nothing at this level   
  updateDCS();
}

void NavBase::reset() { // set to initial conditions (from startup)
  loadIdentity(); // do I really want to do this?
}
//==========>NavModel<====================================
void NavModel::rotate(float rads,char axis){
  switch (axis){
  case 'x':
  case 'X':
    rot[0]=rads;
    break;
  case 'y':
  case 'Y':
    rot[1]=rads;
    break;
  case 'z':
  case 'Z':
    rot[2]=rads;
    break;
  default:
    break;
  }
}
void NavModel::rotate(float *r){
  //rotate(r[0],'x');
  //rotate(r[1],'y');
  //rotate(r[2],'z');
  //navDCS->setRot(r[0],r[1],r[2]);
  rot[0]=r[0];
  rot[1]=r[1];
  rot[2]=r[2];
}
void NavModel::translate(float x,float y,float z){
  // navDCS->setTrans(v[0],v[1],v[2]);
  /* Why am I implementing it this way???
     Because the behavior needs to match that of the 
     NavBase class in order to keep confusion down
     to a minimum.  Also, the setRot() is an absolute
     set and we need to be able to set x,y, and z
     rotation in separate calls in order to be 
     consistent with the NavBase/CaveNav way of
     doing things.

     If you don't like it, then just set the DCS 
     yourself!!! ;)
     */
  trans[0]=x;
  trans[1]=y;
  trans[2]=z; 
}
void NavModel::translate(float *v){
  //navDCS->setTrans(v[0],v[1],v[2]); // absolute.. not relative
  trans[0]=v[0];
  trans[1]=v[1];
  trans[2]=v[2];
}

void NavModel::loadIdentity(){
  pfMatrix m;
  navDCS->getMat(m);
  m.makeIdent();
  navDCS->setMat(m);
  //navDCS->getMatPtr()->makeIdent();
}
  
NavModel::NavModel(pfDCS *model):NavBase(model){
}

void NavModel::updateDCS() {// do nothing at this level  
  navDCS->setRot(rot[0],rot[1],rot[2]);
  navDCS->setTrans(trans[0],trans[1],trans[2]);
  rot[0]=trans[0]=0.0f;
  rot[1]=trans[1]=0.0f;
  rot[2]=trans[2]=0.0f;
}

void NavModel::update() {// do nothing at this level   
  updateDCS();
}

void NavModel::reset() { // set to initial conditions (from startup)
  loadIdentity(); // do I really want to do this?
}
//==========>NavRate<====================================
void NavRate::rotationVelocity(float rads_sec,char axis){
  rotate(dt*rads_sec,axis);
}  
void NavRate::translationVelocity(float vx,float vy,float vz){
  translate(dt*vx,dt*vy,dt*vz);
}
void NavRate::updateTime(){
  //timeval tm;
  // gettimeofday(&tm);
  lastTime=thisTime;
  //thisTime=tm.tv_usec;
  thisTime=CAVEGetTime();
  //dt=(float)(thisTime-lastTime);
  dt=thisTime-lastTime;
  //cerr << "this= " << thisTime << "\tlast= " << lastTime << endl;
  //cerr << "dt= " << dt << endl;
}

NavRate::NavRate(pfDCS *dcs):NavBase(dcs){
  updateTime();
  dt=0.0;
}

void NavRate::update(){
  updateTime();
  //pfCAVEDCSNavTransform(navDCS);
}

void NavRate::reset(){
  updateTime();
  dt=0;
  lastTime=thisTime;
  loadIdentity(); // do I really want to do this?
}

//=========>NavFly<======================================

NavFly::NavFly(pfDCS *nav):NavRate(nav){
  rSpeed=DEFAULT_RSPEED;
  tSpeed=DEFAULT_TSPEED;
}

void NavFly::update(){
  float w[3];
  float jx,jy;
  updateTime();
  CaveWand.unitVectorDirection(w);
  jx=CAVE_JOYSTICK_X;
  jy=CAVE_JOYSTICK_Y;
  // y=translate in wand direction
  if(THRESH(jy,0.1))
    translationVelocity(w[0]*jy*tSpeed,w[1]*jy*tSpeed,w[2]*jy*tSpeed);
  // x=rotate left/right on z axis
  if(THRESH(jx,0.2))
    rotationVelocity(-rSpeed*jx,'z');  
  //pfCAVEDCSNavTransform(navDCS); // update DCS
  // NavBase::update();
}

void NavFly::setRotationSpeed(float speed){
  // where speed is in rad/sec
  rSpeed=speed;
}

void NavFly::adjustRotationSpeed(float factor){
  rSpeed*=factor;
}

void NavFly::setTranslationSpeed(float speed){
  // where speed is the trans speed in clicks/sec
  tSpeed=speed;
}

void NavFly::adjustTranslationSpeed(float factor){
  tSpeed*=factor;
}


// class NavGrab??  grab() ungrab()
//================>NavGrabAbsolute<====================
void NavGrabAbsolute::negateVF3(float *v){
  register int i;
  for(i=0;i<3;i++) v[i]=-v[i];
}
NavGrabAbsolute::NavGrabAbsolute(pfDCS *dcs):NavBase(dcs){
  update(); 	
}

void NavGrabAbsolute::update(){
  float w[3];
  // start by snapping to front of wand
  loadIdentity();
  // snap to wand
  CaveWand.orientation(w);
  negateVF3(w);
  rotate(w);
  CaveWand.position(w);
  negateVF3(w);
  translate(w);
  // pfCAVEDCSNavTransform(navDCS);
  // NavBase::update();
}

//================>NavGrabAbsolute<====================
void NavGrabRelative::offsetV3(float initial[3],float current[3],float offset[3]){
  register int i;
  for(i=0;i<3;i++) offset[i]=-(initial[i]+current[i]);
}
void NavGrabRelative::negateVF3(float *v){
  register int i;
  for(i=0;i<3;i++) v[i]=-v[i];
}
NavGrabRelative::NavGrabRelative(pfDCS *dcs):NavBase(dcs){
  // start by recording wand initial position
  CaveWand.orientation(initialWandOrientation);
  CaveWand.position(initialWandPosition);
}
void NavGrabRelative::update(){
  float wandOrientationOffset[3],w[3];
  float wandPositionOffset[3];
    
  CaveWand.orientation(w);
  offsetV3(initialWandOrientation,w,wandOrientationOffset);
    
  CaveWand.position(w);
  offsetV3(initialWandPosition,w,wandPositionOffset);
    
  // now set the DCS with this info
  loadIdentity();
  rotate(wandOrientationOffset);
  translate(wandPositionOffset);
  //pfCAVEDCSNavTransform(navDCS);
  //NavBase::update();
}
void NavGrabRelative::reset(){
  CaveWand.orientation(initialWandOrientation);
  CaveWand.position(initialWandPosition);
  loadIdentity();
}

//==========>NavWalk<=================================
NavWalk::NavWalk(pfDCS *navdcs,pfNodePtr &mdl,boolean icache):
  NavFly(navdcs),model(mdl) { 
  segset.activeMask = 1;
  segset.isectMask = FLOORMASK; // only intersect with floor geom
  //segset.isectMask = 0xFFFF;
  segset.discFunc = NULL;
  segset.bound = NULL;
  segset.mode = PFTRAV_IS_PRIM;
  segset.segs[0].dir.set(0.0f, 0.0f, -1.0f);
  segset.segs[0].length = 1000.0f;

  segset.segs[1].dir.set(0.0f,0.0f,1.0f); // test upwards
  segset.segs[1].length = 1000.0f;

  samplerate=1; nsamples=0;
  velocity=acceleration=0;

  if(icache){
    // Don't attempt to modify mask... just make certain it is cached..
    model->setTravMask(PFTRAV_ISECT,0xffffffff,
		       PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE,PF_AND);
  }
}

void NavWalk::computeNavHeight(){
  float head[3],worldHead[3];  
  pfHit **hits[32];
  int isect;
  int segnum=0; // initially test downwards
  //double thisTime=CAVEGetTime();
  //double dt=thisTime-lastTime;
  

  //printf ("computing nav height: \n");
  CaveHead.position(head);
  CAVENavConvertCAVEToWorld(head,worldHead);
  segset.segs[0].pos.set(worldHead[0],worldHead[1],worldHead[2]);
  isect = model->isect(&segset, hits);
  if(!isect){ // test upwards
    segnum=1;
    //cerr << "Test Upwards" << endl;
    segset.activeMask=2; // use upward segment to test
    segset.segs[1].pos.set(worldHead[0],worldHead[1],worldHead[2]);
    isect = model->isect(&segset, hits);
    segset.activeMask=1; // return to normal testmode
  }
  if (isect){
    pfVec3 pnt, xpnt;
    pfMatrix xmat;
    float foot[3],worldFoot[3];
    foot[0] = head[0];
    foot[1] = head[1];
    foot[2] = 0;  // set flyheight later
    CAVENavConvertCAVEToWorld(foot,worldFoot);
    (*hits[segnum])->query(PFQHIT_POINT, pnt.vec);
    (*hits[segnum])->query(PFQHIT_XFORM, (float*)xmat.mat);
    xpnt.xformPt(pnt, xmat);
    translate(0,0,xpnt[PF_Z]-worldFoot[2]);
    // compute velocity and acceleration
    //double newVel,newAcc;
    // need a measure of distance from last time
    //lastFoot=worldFoot[2];
    //printf ("%f\n", worldFoot[2]);
  }
  /*
  else{
    // no ground...
    velocity=acceleration=0;
  }*/
  lastTime=thisTime;
}

void NavWalk::update(){
  NavFly::update();
  if(!((nsamples++)%samplerate))
    computeNavHeight();
  // else
}

void NavWalk::intersectionCaching(boolean icache){
  if(icache){
    model->setTravMask(PFTRAV_ISECT,0xffffffff,
		     PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_CACHE,PF_AND);
  }
  else {
    model->setTravMask(PFTRAV_ISECT,0xffffffff,
		       PFTRAV_SELF|PFTRAV_DESCEND|PFTRAV_IS_UNCACHE,PF_AND);
  } 
}

