//------Standard C/C++ libraries
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

//------John's Templates
#include "FlexArrayTmpl.H"

//------Header
#include "ReadAtomFiles.hh"

AtomType::AtomType(AtomType &at){
  for(int i=0;i<atoms.getSize();i++)
    getTypeNumber(at.atoms[i].name);
}
int AtomType::getTypeNumber(char *name){
  for(int i=0;i<atoms.getSize();i++)
    if(!strcmp(atoms[i].name,name))
      return i;
  //printf("Add new atomtype for %s\n",name);
  atoms.setSize(atoms.getSize()+1); // increase size
  strcpy((atoms[atoms.getSize()-1]).name,name);
  return (atoms.getSize()-1);
}
AtomType &AtomType::operator=(AtomType &at){
  if(&at==this) return *this;
  //puts("Copy Atom Types... have %u types\n",at.nTypes());
  for(int i=0;i<at.nTypes();i++)
    getTypeNumber(at.atoms[i].name);
  return *this;
}

// copy constructor (for copying Atom Lists)
AtomList::AtomList(AtomList &a){
  atoms = a.atoms;
}
// copy operator (for copying Atom Lists)
AtomList &AtomList::operator=(AtomList &a){
  if(&a!=this)
    atoms=a.atoms;
  return *this;
}
AtomList::AtomList(char *filename){
  read(filename);
}
void AtomList::read(char *filename){
  char buffer[1024];
  FILE *file=fopen(filename,"r");
  if(!file){
    printf("AtomList::read() could not open file %s\n",filename);
    return;
  }
  fgets(buffer,sizeof(buffer)-1,file);
  int len=atoi(buffer);
  fgets(buffer,sizeof(buffer)-1,file);
  while(fgets(buffer,sizeof(buffer)-1,file)){
    AtomDef atom;
    char name[32];
    sscanf(buffer,"%s %f %f %f",
	   name,&(atom.p.cart.x),&(atom.p.cart.y),&(atom.p.cart.z));
    atom.type = atom_types.getTypeNumber(name);
    /* The old hardcoded method below has been replaced by
       the "atom_types" method above.
     if(strstr(name,"Al"))
       atom.type=1;
     else 
      atom.type=2;
      */
    // printf("Appending %s\n",buffer); // debug
    atoms.append(atom);
  }
  fclose(file);
} 

void AtomList::print(){
  printf("Number of records = %u\n",atoms.getSize());
  for(int i=0;i<atoms.getSize();i++){
    printf("Type=%u: pos=%f:%f:%f\n",
	   atoms[i].type,
	   atoms[i].p.cart.x,atoms[i].p.cart.y,atoms[i].p.cart.z);
  }
}
char *TimeAtomList::generateName(char *buffer,
					char *basename,int index){
  sprintf(buffer,"%s.%2u",basename,index);
  return buffer;
}
TimeAtomList::TimeAtomList(char *namebase,int minindex,int maxindex){
  // declare the size of the timesteps array ahead of time
  timesteps.setSize(1+maxindex-minindex);
  printf("setting size to %u\n",1+maxindex-minindex);
  // Read in the data for the timesteps
  for(int idx=0,i=minindex;i<=maxindex;idx++,i++){
    char buffer[128];
    char *filename=generateName(buffer,namebase,i);
    printf("Loading file %s\n",filename);
    //(timesteps[idx]).print();
    (timesteps[idx]).read(filename);
    //(timesteps[idx]).print();
  }
  atom_types = (timesteps[0]).atom_types;
}
// destructor (called when you "delete" this object)
TimeAtomList::~TimeAtomList(){ // nil for now
}
void TimeAtomList::getRange(float &min,float &max){
  min=0;
  max=(float)(timesteps.getSize()-1);
}

// linearly interpolate between timesteps
// based on a floating-point "time" value
// put the results in the array "atoms"
void TimeAtomList::getAtoms(float time,FlexArray<AtomDef> &atoms){
  int timeindex=(int)time;
  if(timeindex==(timesteps.getSize()-1))
    timeindex--;
  float normdistance = time - (float)timeindex;
  // make sure input atom list is the same size
  // as the atom lists it is interpolating
  int natoms=(timesteps[timeindex]).atoms.getSize();
  if(atoms.getSize()!=natoms){
    atoms.setSize(natoms);
    for(int i=0;i<natoms;i++)
      atoms[i].type = timesteps[timeindex].atoms[i].type;
  }
  for(int i=0;i<natoms;i++){
    // just pointing to arrays that are deeply 
    // nested in data structures (just for
    // convenience)
    float *start=timesteps[timeindex].atoms[i].p.array;
    float *end=timesteps[timeindex+1].atoms[i].p.array;
    float *interp=atoms[i].p.array;
    // and here is the linear interpolation
    for(int j=0;j<3;j++)
      interp[j] = start[j] + (end[j]-start[j])*normdistance;
  }
}

void print(FlexArray<AtomDef> &atoms){
  printf("Number of records = %u\n",atoms.getSize());
  for(int i=0;i<10;i++){
    printf("Type=%u: pos=%f:%f:%f\n",
	   atoms[i].type,
	   atoms[i].p.cart.x,atoms[i].p.cart.y,atoms[i].p.cart.z);
  }
}
